import anndata
import numpy as np
import pandas as pd
import pytest
import scanpy as sc
from numpy.random import negative_binomial, binomial, seed
from scipy import sparse as sp

from cirrocumulus.de import DE


def get_example_data(sparse=False):
    seed(1234)
    # create test object
    adata = anndata.AnnData(
        np.multiply(binomial(1, 0.15, (100, 20)), negative_binomial(2, 0.25, (100, 20)))
    )
    # adapt marker_genes for cluster (so as to have some form of reasonable input
    adata.X[0:10, 0:5] = np.multiply(
        binomial(1, 0.9, (10, 5)), negative_binomial(1, 0.5, (10, 5))
    )

    # The following construction is inefficient, but makes sure that the same data is used in the sparse case
    if sparse:
        adata.X = sp.csr_matrix(adata.X)

    # Create cluster according to groups
    adata.obs['sc_groups'] = pd.Categorical(
        np.concatenate(
            (
                np.zeros((10,), dtype=int),
                np.ones((90,), dtype=int),
            )
        )
    )

    return adata


@pytest.fixture(autouse=True, params=[True, False])
def sparse(request):
    return request.param


def test_de(sparse):
    adata = get_example_data(sparse)

    batch_size = 3
    obs_field = 'sc_groups'
    nfeatures = adata.shape[1]
    get_batch_fn = lambda i: adata[:, i:min(nfeatures, i + batch_size)]
    de = DE(adata, obs_field, nfeatures, batch_size, get_batch_fn)

    sc.tl.rank_genes_groups(adata, obs_field, method='t-test', pts=True)
    rank_genes_groups = adata.uns['rank_genes_groups']
    sc_scores = rank_genes_groups['scores']['0']
    sc_pvals = rank_genes_groups['pvals']['0']
    sc_pts = rank_genes_groups['pts']
    sc_lfc = rank_genes_groups['logfoldchanges']['0']

    features = rank_genes_groups['names']['0']
    sc_df = pd.DataFrame(index=features,
                         data={'pvals': sc_pvals, 'scores': sc_scores, 'pts': sc_pts['0'], 'pts_rest': sc_pts['1'],
                               'lfc': sc_lfc})
    sc_df = sc_df.loc[adata.var.index]
    np.testing.assert_allclose(sc_df['scores'], de.scores)
    np.testing.assert_allclose(sc_df['pvals'], de.pvals)
    np.testing.assert_allclose(sc_df['lfc'], de.logfoldchanges)
    if sparse:
        np.testing.assert_allclose(sc_df['pts'], de.frac_expressed.iloc[0])
        np.testing.assert_allclose(sc_df['pts_rest'], de.frac_expressed.iloc[1])
    else:
        assert de.frac_expressed is None




# import pandas.io.json as json
from urllib.parse import urlparse

import fsspec
import pandas._libs.json as ujson
from flask import make_response


def get_scheme(path):
    pr = urlparse(path)
    if len(pr.scheme) <= 1:  # for file paths: /foo/bar/test.h5ad or C:/foo/bar/test.h5ad
        return 'file'
    return pr.scheme


scheme_to_fs = {}


def get_fs(path):
    scheme = get_scheme(path)
    fs = scheme_to_fs.get(scheme, None)
    if fs is not None:
        return fs
    fs = fsspec.filesystem(scheme)
    scheme_to_fs[scheme] = fs
    return fs


def to_json(data):
    return ujson.dumps(data, double_precision=2, orient='values')


def json_response(data, response=200):
    # response = make_response(simplejson.dumps(data, check_circular=True), response)
    # response = make_response(json.dumps(data), response)
    s = ujson.dumps(data, double_precision=2, orient='values')
    # s = nujson.dumps(data, double_precision=1)
    response = make_response(s, response)
    response.headers['Content-Type'] = 'application/json'
    return response


def get_email_domain(email):
    at_index = email.find('@')
    domain = None
    if at_index != -1:
        domain = email[at_index + 1:]
    return domain


def load_dataset_schema(url):
    import os

    import fsspec
    import json

    def get_extension(path):
        name, ext = os.path.splitext(path)
        if ext == '.gz':
            name, ext = os.path.splitext(name)
            if ext == '.json':
                ext = '.json.gz'
        return ext

    extension = get_extension(url)
    json_schema = None
    if extension in ['.json', '.json.gz', '']:
        scheme = get_scheme(url)
        fs = fsspec.filesystem(scheme)
        if extension == '':
            url = os.path.join(url, 'index.json.gz')
            extension = get_extension(url)
        if extension == '.json.gz':
            import gzip
            with gzip.open(fs.open(url)) as f:
                json_schema = json.load(f)
        else:
            with fs.open(url) as f:
                json_schema = json.load(f)
    return json_schema


def write_top_half_gct(f, row_metadata_df, col_metadata_df, metadata_null, filler_null):
    """ Write the top half of the gct file: top-left filler values, row metadata
    headers, and top-right column metadata.
    Args:
        f (file handle): handle for output file
        row_metadata_df (pandas df)
        col_metadata_df (pandas df)
        metadata_null (string): how to represent missing values in the metadata
        filler_null (string): what value to fill the top-left filler block with
    Returns:
        None
    """
    # Initialize the top half of the gct including the third line
    size_of_top_half_df = (1 + col_metadata_df.shape[1],
                           1 + row_metadata_df.shape[1] + col_metadata_df.shape[0])

    top_half_df = pd.DataFrame(np.full(size_of_top_half_df, filler_null, dtype=object))

    # Assemble the third line of the gct: "id", then rhds, then cids
    top_half_df.iloc[0, :] = np.hstack(("id", row_metadata_df.columns.values, col_metadata_df.index.values))

    # Insert the chds
    top_half_df.iloc[range(1, top_half_df.shape[0]), 0] = col_metadata_df.columns.values

    # Insert the column metadata, but first convert to strings and replace NaNs
    col_metadata_indices = (range(1, top_half_df.shape[0]),
                            range(1 + row_metadata_df.shape[1], top_half_df.shape[1]))
    # pd.DataFrame.at to insert into dataframe(python3)
    top_half_df.at[col_metadata_indices[0], col_metadata_indices[1]] = (
        col_metadata_df.astype(str).replace("nan", value=metadata_null).T.values)

    # Write top_half_df to file
    top_half_df.to_csv(f, header=False, index=False, sep="\t")


def write_bottom_half_gct(f, row_metadata_df, data_df, data_null, data_float_format, metadata_null):
    """ Write the bottom half of the gct file: row metadata and data.
    Args:
        f (file handle): handle for output file
        row_metadata_df (pandas df)
        data_df (pandas df)
        data_null (string): how to represent missing values in the data
        metadata_null (string): how to represent missing values in the metadata
        data_float_format (string): how many decimal points to keep in representing data
    Returns:
        None
    """
    # create the left side of the bottom half of the gct (for the row metadata)
    size_of_left_bottom_half_df = (row_metadata_df.shape[0],
                                   1 + row_metadata_df.shape[1])
    left_bottom_half_df = pd.DataFrame(np.full(size_of_left_bottom_half_df, metadata_null, dtype=object))

    # create the full bottom half by combining with the above with the matrix data
    bottom_half_df = pd.concat([left_bottom_half_df, data_df.reset_index(drop=True)], axis=1)
    bottom_half_df.columns = range(bottom_half_df.shape[1])

    # Insert the rids
    bottom_half_df.iloc[:, 0] = row_metadata_df.index.values

    # Insert the row metadata, but first convert to strings and replace NaNs
    row_metadata_col_indices = range(1, 1 + row_metadata_df.shape[1])
    bottom_half_df.iloc[:, row_metadata_col_indices] = (
        row_metadata_df.astype(str).replace("nan", value=metadata_null).values)

    # Write bottom_half_df to file
    bottom_half_df.to_csv(f, header=False, index=False, sep="\t",
                          na_rep=data_null,
                          float_format=data_float_format)


def adata2gct(adata, f):
    data_float_format = "%.4f"
    f.write("#1.3\n")
    f.write(str(adata.shape[0]) + "\t" + str(adata.shape[1]) + "\t" + str(adata.obs.shape[1]) + "\t" + str(
        adata.var.shape[1]) + "\n")
    write_top_half_gct(f, adata.obs, adata.var, '', '')
    write_bottom_half_gct(f, adata.obs, pd.DataFrame(adata.X), '', data_float_format, '')



import os

import numpy as np
import pandas as pd
from anndata import AnnData
from cirrocumulus.de import DE

from .data_processing import get_filter_str, get_mask
from .diff_exp import fdrcorrection
from .envir import CIRRO_SERVE, CIRRO_MAX_WORKERS

executor = None


def submit_job(database_api, dataset_api, email, dataset, job_name, job_type, params):
    global executor
    if executor is None:
        from concurrent.futures.thread import ThreadPoolExecutor
        import os

        if os.environ.get(CIRRO_SERVE) != 'true':
            max_workers = 1
        else:
            max_workers = int(os.environ.get(CIRRO_MAX_WORKERS, '2'))
        executor = ThreadPoolExecutor(max_workers=max_workers)
    job_id = database_api.create_job(email=email, dataset_id=dataset['id'], job_name=job_name, job_type=job_type,
                                     params=params)
    executor.submit(run_job, database_api, dataset_api, email, job_id, job_type, dataset, params)
    return job_id


def _power(X, power):
    return X ** power if isinstance(X, np.ndarray) else X.power(power)


def run_job(database_api, dataset_api, email, job_id, job_type, dataset, params):
    database_api.update_job(email=email, job_id=job_id, status='running', result=None)
    schema = dataset_api.schema(dataset)
    var_names = schema['var']
    nfeatures = len(var_names)
    filters = [params['filter']]
    if job_type == 'de':
        filters.append(params['filter2'])
    masks, _ = get_mask(dataset_api, dataset, filters)
    batch_size = 5000 if os.environ.get(CIRRO_SERVE) == 'true' else nfeatures  # TODO more intelligent batches
    obs_field = 'tmp'

    def get_batch_fn(i):
        start = i
        end = min(nfeatures, start + batch_size)
        features = var_names[start:end]
        adata = dataset_api.read_dataset(keys=dict(X=features), dataset=dataset)
        if batch_size != nfeatures:
            frac = end / nfeatures
            database_api.update_job(email=email, job_id=job_id,
                                    status='running {:.0f}%'.format(100 * frac) if frac < 1 else 'saving results',
                                    result=None)
        return adata

    obs = pd.DataFrame(index=pd.RangeIndex(schema['shape'][0]))
    obs[obs_field] = ''
    for i in range(len(masks)):
        obs.loc[masks[i], obs_field] = str(i)
    obs[obs_field] = obs[obs_field].astype('category')
    de = DE(AnnData(obs=obs), obs_field, nfeatures, batch_size, get_batch_fn,
            key_set=[str(i) for i in range(len(masks))])
    pvals = fdrcorrection(de.pvals)
    # group:field is object entry
    if job_type == 'de':
        comparison1_name = get_filter_str(params['filter'])
    comparison2_name = get_filter_str(params['filter2'])
    comparison = 'comparison' if comparison1_name is None or comparison2_name is None else comparison1_name + '_' + comparison2_name

    result_df = pd.DataFrame(
        data={'index': var_names, '{}:pvals_adj'.format(comparison): pvals, '{}:scores'.format(comparison): de.scores,
              '{}:lfc'.format(comparison): de.logfoldchanges})
    if de.frac_expressed is not None:
        result_df['{}:pts_1'.format(comparison)] = de.frac_expressed.iloc[0]
        result_df['{}:pts_2'.format(comparison)] = de.frac_expressed.iloc[1]
    result = dict(groups=[comparison],
                  format='json',
                  fields=['pvals_adj', 'scores', 'lfc'] + (['pts_1', 'pts_2'] if de.frac_expressed is not None else []),
                  data=result_df.to_dict(orient='records'))
    database_api.update_job(email=email, job_id=job_id, status='complete', result=result)



import numpy as np
import pandas as pd
import scipy.stats

from cirrocumulus.groupby import GroupBy


class DE:

    def __init__(self, adata, obs_field, nfeatures, batch_size, get_batch_fn, key_set=None):
        group_by = GroupBy(adata, obs_field, key_set=key_set)
        A, keys = group_by.sparse_aggregator(True)
        mean = None
        variance = None
        count = None
        frac_expressed = None
        for i in range(0, nfeatures, batch_size):
            group_by.adata = get_batch_fn(i)  # hack to update anndata in groupby

            result = group_by.count_mean_var(A, keys)
            # groups on rows, genes on columns
            mean = pd.concat((mean, result['mean']), axis=1) if mean is not None else result['mean']
            variance = pd.concat((variance, result['var']), axis=1) if variance is not None else result['var']
            if count is None:
                count = result['count']
            if result['frac_expressed'] is not None:
                frac_expressed = pd.concat((frac_expressed, result['frac_expressed']),
                                           axis=1) if frac_expressed is not None else result['frac_expressed']
        mean1 = mean.iloc[0]
        mean2 = mean.iloc[1]
        if 'log1p' in adata.uns_keys() and adata.uns['log1p']['base'] is not None:
            expm1_func = lambda x: np.expm1(x * np.log(adata.uns['log1p']['base']))
        else:
            expm1_func = np.expm1
        # foldchanges = (expm1_func(mean1) + 1e-9) / (expm1_func(mean2) + 1e-9)  # add small value to remove 0's
        foldchanges = np.log2((expm1_func(mean1) + 1e-9) / (expm1_func(mean2) + 1e-9))
        variance1 = variance.iloc[0]
        variance2 = variance.iloc[1]
        nobs1 = count.iloc[0]
        nobs2 = count.iloc[1]
        with np.errstate(invalid="ignore"):
            scores, pvals = scipy.stats.ttest_ind_from_stats(
                mean1=mean1,
                std1=np.sqrt(variance1),
                nobs1=nobs1,
                mean2=mean2,
                std2=np.sqrt(variance2),
                nobs2=nobs2,
                equal_var=False,  # Welch's
            )

        scores[np.isnan(scores)] = 0
        pvals[np.isnan(pvals)] = 1
        self.scores = scores
        self.pvals = pvals
        self.logfoldchanges = foldchanges
        self.frac_expressed = frac_expressed



CIRRO_AUTH_CLIENT_ID = 'CIRRO_AUTH_CLIENT_ID'
CIRRO_DB_URI = 'CIRRO_DB_URI'
CIRRO_EMAIL = 'CIRRO_EMAIL'
CIRRO_SERVE = 'CIRRO_SERVE'
CIRRO_MAX_WORKERS = 'CIRRO_MAX_WORKERS'
CIRRO_FOOTER = 'CIRRO_FOOTER'
CIRRO_BRAND = 'CIRRO_BRAND'
CIRRO_UPLOAD = 'CIRRO_UPLOAD'
CIRRO_DATABASE_CLASS = 'CIRRO_DATABASE'

SERVER_CAPABILITY_RENAME_CATEGORIES = 'rename_categories'
SERVER_CAPABILITY_JOBS = 'jobs'
SERVER_CAPABILITY_SAVE_FEATURE_SETS = 'save_feature_sets'
SERVER_CAPABILITY_SAVE_LINKS = 'save_links'
SERVER_CAPABILITY_EDIT_DATASET = 'edit_dataset'
SERVER_CAPABILITY_ADD_DATASET = 'add_dataset'
SERVER_CAPABILITY_DELETE_DATASET = 'delete_dataset'



from enum import Enum

import numpy as np
import pandas as pd
from cirrocumulus.io_util import cirro_id


class DataType(Enum):
    UNSPECIFIED = 'unspecified'
    MODULE = 'module'


def get_scanpy_marker_keys(dataset):
    scanpy_marker_keys = []
    for key in dataset.uns.keys():
        rank_genes_groups = dataset.uns[key]
        if isinstance(rank_genes_groups, dict) and 'names' in rank_genes_groups and (
                'pvals' in rank_genes_groups or 'pvals_adj' in rank_genes_groups or 'scores' in rank_genes_groups):
            scanpy_marker_keys.append(key)
    return scanpy_marker_keys


def get_pegasus_marker_keys(dataset):
    marker_keys = []
    for key in dataset.varm.keys():
        d = dataset.varm[key]
        if isinstance(d, np.recarray):
            try:
                ''.join(d.dtype.names).index('log2Mean')
                marker_keys.append(key)
            except ValueError:
                pass
    return marker_keys


def obs_stats(adata, columns):
    df = adata.obs[columns]
    # variables on columns, stats on rows, transpose so that stats are on columns
    return df.agg(['min', 'max', 'sum', 'mean']).T


def X_stats(adata):
    X = adata.X
    return pd.DataFrame(
        data={'min': X.min(axis=0).toarray().flatten(),
              'max': X.max(axis=0).toarray().flatten(),
              'sum': X.sum(axis=0).flatten(),
              'numExpressed': X.getnnz(axis=0),
              'mean': X.mean(axis=0)}, index=adata.var.index)


def datasets_schema(datasets):
    """ Gets dataset schema.

    Returns
        schema dict. Example:
        {"version":"1.0.0",
        "categoryOrder":{
            "louvain":["0","1","2","3","4","5","6","7"],
            "leiden":["0","1","2","3","4","5","6","7"]},
        "var":["TNFRSF4","CPSF3L","ATAD3C"],
        "obs":["percent_mito","n_counts"],
        "obsCat":["louvain","leiden"],
        "shape":[2638,1838],
        "embeddings":[{"name":"X_pca","dimensions":3},{"name":"X_pca","dimensions":2},{"name":"X_umap","dimensions":2}]
    }
    """
    obs_cat = []
    obs = []

    marker_results = []
    de_results_format = 'records'
    for dataset in datasets:
        prior_marker_results = dataset.uns.get('markers', [])

        if isinstance(prior_marker_results, str):
            import json
            prior_marker_results = json.loads(prior_marker_results)
    marker_results += prior_marker_results

    n_genes = 10
    de_results = []  # array of dicts containing params logfoldchanges, pvals_adj, scores, names
    category_to_order = {}
    embeddings = []
    field_to_value_to_color = dict()  # field -> value -> color
    for i in range(len(datasets)):
        dataset = datasets[i]
        for scanpy_marker_key in get_scanpy_marker_keys(dataset):
            rank_genes_groups = dataset.uns[scanpy_marker_key]
            has_fc = 'logfoldchanges' in rank_genes_groups
            min_fold_change = 1
            params = rank_genes_groups['params']
            prefix = None if i == 0 else dataset.uns.get('name')

            # pts and pts_rest in later scanpy versions
            rank_genes_groups_keys = list(rank_genes_groups.keys())
            for k in ['params', 'names']:
                if k in rank_genes_groups_keys:
                    rank_genes_groups_keys.remove(k)
            if 'pvals' in rank_genes_groups_keys and 'pvals_adj' in rank_genes_groups_keys:
                rank_genes_groups_keys.remove('pvals')
            category = '{} ({})'.format(params['groupby'], scanpy_marker_key)
            de_result_name = category if prefix is None else prefix + '-' + category
            de_result_df = None
            group_names = rank_genes_groups['names'].dtype.names
            for group_name in group_names:
                group_df = pd.DataFrame(index=rank_genes_groups['names'][group_name])
                group_df = group_df[group_df.index != 'nan']
                for rank_genes_groups_key in rank_genes_groups_keys:
                    values = rank_genes_groups[rank_genes_groups_key][group_name]
                    column_name = '{}:{}'.format(group_name, rank_genes_groups_key)
                    group_df[column_name] = values

                if de_result_df is None:
                    de_result_df = group_df
                else:
                    de_result_df = de_result_df.join(group_df, how='outer')
                if n_genes > 0:
                    markers_df = group_df
                    if has_fc:
                        markers_df = group_df[group_df['{}:logfoldchanges'.format(group_name)] > min_fold_change]
                    if len(markers_df) > 0:
                        marker_results.append(
                            dict(category=de_result_name, name=str(group_name), features=markers_df.index[:n_genes]))

            if de_results_format == 'records':
                de_result_data = de_result_df.reset_index().to_dict(orient='records')
            else:
                de_result_data = dict(index=de_result_df.index)
                for c in de_result_df:
                    de_result_data[c] = de_result_df[c]

            de_result = dict(id=cirro_id(),
                             color='logfoldchanges' if 'logfoldchanges' in rank_genes_groups_keys else
                             rank_genes_groups_keys[0],
                             size='pvals_adj' if 'pvals_adj' in rank_genes_groups_keys else rank_genes_groups_keys[
                                 0],
                             params=params, groups=group_names, fields=rank_genes_groups_keys, type='de',
                             name=de_result_name)
            de_result['data'] = de_result_data
            de_results.append(de_result)
        for varm_field in get_pegasus_marker_keys(dataset):
            de_res = dataset.varm[varm_field]
            names = de_res.dtype.names
            field_names = set()  # e.g. 1:auroc
            group_names = set()
            for name in names:
                index = name.rindex(':')
                field_name = name[index + 1:]
                group_name = name[:index]
                field_names.add(field_name)
                group_names.add(group_name)
            group_names = list(group_names)
            field_names = list(field_names)

            de_result_df = pd.DataFrame(data=de_res, index=dataset.var.index)
            de_result_df.index.name = 'index'
            if de_results_format == 'records':
                de_result_data = de_result_df.reset_index().to_dict(orient='records')
            else:
                de_result_data = dict(index=de_result_df.index)
                for c in de_res:
                    de_result_data[c] = de_result_df[c]

            de_result = dict(id=cirro_id(), type='de', name='pegasus_de',
                             color='log2FC' if 'log2FC' in field_names else field_names[0],
                             size='mwu_qval' if 'mwu_qval' in field_names else field_names[0],
                             groups=group_names,
                             fields=field_names)
            de_result['data'] = de_result_data
            de_results.append(de_result)

            if n_genes > 0:
                field_use = None
                for field in ['mwu_qval', 'auroc', 't_qval']:
                    if field in field_names:
                        field_use = field
                        break
                if field_use is not None:
                    field_ascending = field_use != 'auroc'
                    for group_name in group_names:
                        fc_column = '{}:log2FC'.format(group_name)
                        name = '{}:{}'.format(group_name, field_name)

                        idx_up = de_result_df[fc_column] > 0
                        df_up = de_result_df.loc[idx_up].sort_values(by=[name, fc_column],
                                                                     ascending=[field_ascending, False])
                        features = df_up[:n_genes].index.values
                        marker_results.append(dict(category='markers', name=str(group_name), features=features))

        for key in dataset.obs_keys():
            if pd.api.types.is_categorical_dtype(dataset.obs[key]) or pd.api.types.is_bool_dtype(
                    dataset.obs[key]) or pd.api.types.is_object_dtype(dataset.obs[key]):
                obs_cat.append(key)
            else:
                obs.append(key)

        # if 'metagenes' in dataset.uns:
        #     metagenes = dataset.uns['metagenes']
        #     schema_dict['metafeatures'] = metagenes['var'].index

        for key in dataset.obs_keys():
            if pd.api.types.is_categorical_dtype(dataset.obs[key]) and len(dataset.obs[key]) < 1000:
                category_to_order[key] = dataset.obs[key].cat.categories

        # spatial_node = adata.uns['spatial'] if 'spatial' in adata.uns else None
        #
        # if spatial_node is not None:
        #     spatial_node_keys = list(spatial_node.keys())  # list of datasets
        #     if len(spatial_node_keys) == 1:
        #         spatial_node = spatial_node[spatial_node_keys[0]]

        images_node = dataset.uns.get('images',
                                      [])  # list of {type:image or meta_image, name:image name, image:path to image, spot_diameter:Number}
        image_names = list(map(lambda x: x['name'], images_node))

        for key in dataset.obsm_keys():
            dim = dataset.obsm[key].shape[1]
            if 1 < dim <= 3:
                embedding = dict(name=key, dimensions=dim)
                if dim == 2:
                    try:
                        image_index = image_names.index(key)
                        embedding['spatial'] = images_node[image_index]
                    except ValueError:
                        pass

                embeddings.append(embedding)
        meta_images = dataset.uns.get('meta_images', [])
        for meta_image in meta_images:
            embeddings.append(meta_image)

        for key in dataset.uns.keys():
            if key.endswith('_colors'):
                field = key[0:len(key) - len('_colors')]
                if field in dataset.obs:
                    colors = dataset.uns[key]
                    if pd.api.types.is_categorical_dtype(dataset.obs[field]):
                        categories = dataset.obs[field].cat.categories
                        if len(categories) == len(colors):
                            color_map = dict()
                            for i in range(len(categories)):
                                color_map[str(categories[i])] = colors[i]
                            field_to_value_to_color[field] = color_map
                    else:
                        print("Skipping colors for {}".format(key))
    schema_dict = {'version': '1.0.0'}
    schema_dict['results'] = de_results
    schema_dict['colors'] = field_to_value_to_color
    schema_dict['markers'] = marker_results
    schema_dict['embeddings'] = embeddings
    schema_dict['categoryOrder'] = category_to_order
    var_ids = []
    modules = None
    for dataset in datasets:
        if dataset.uns.get('data_type', '') == DataType.MODULE:
            modules = dataset.var if modules is None else pd.concat((modules, dataset.var))
        else:
            var_ids += dataset.var.index.to_list()
    if modules is not None:
        modules.index.name = 'id'
        schema_dict['modules'] = modules.reset_index().to_dict(orient='records')
    schema_dict['var'] = var_ids
    schema_dict['obs'] = obs
    schema_dict['obsCat'] = obs_cat

    schema_dict['shape'] = [datasets[0].shape[0], len(var_ids)]
    return schema_dict



import json
from abc import abstractmethod, ABC


class AbstractDataset(ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_suffixes(self):
        pass

    @abstractmethod
    def read_dataset(self, filesystem, path, keys=None, dataset=None):
        pass

    def schema(self, filesystem, path):
        if path.endswith('.gz'):
            import gzip
            with gzip.open(filesystem.open(path)) as s:
                return json.load(s)
        else:
            with filesystem.open(path) as s:
                return json.load(s)



import json
import os

from cirrocumulus.abstract_db import AbstractDB
from cirrocumulus.envir import *
from cirrocumulus.io_util import unique_id


def create_dataset_meta(path):
    result = {'id': path, 'url': path, 'name': os.path.splitext(os.path.basename(path))[0], 'description': ''}
    if os.path.basename(path).endswith('.json'):
        with open(path, 'rt') as f:
            result.update(json.load(f))

    return result


def write_json(json_data, json_path):
    if os.path.exists(os.path.dirname(os.path.abspath(json_path))):  # only support writing local files
        with open(json_path, 'wt') as f:
            json.dump(json_data, f)
    else:
        print('Skipping {}'.format(json_path))


class LocalDbAPI(AbstractDB):

    def __init__(self, paths):
        super().__init__()
        self.dataset_to_info = {}  # json_data, meta, json_path

        for path in paths:
            json_data = {}
            basename = os.path.splitext(path)[0]
            old_path = basename + '_filters.json'
            json_path = basename + '.json'
            if os.path.exists(old_path) and os.path.getsize(old_path) > 0:
                with open(old_path, 'rt') as f:
                    json_data['filters'] = json.load(f)

            if os.path.exists(json_path) and os.path.getsize(json_path) > 0:
                with open(json_path, 'rt') as f:
                    json_data.update(json.load(f))
            meta = create_dataset_meta(path)
            if 'filters' not in json_data:
                json_data['filters'] = {}
            if 'views' not in json_data:
                json_data['views'] = {}
            if 'categories' not in json_data:
                json_data['categories'] = {}
            self.dataset_to_info[path] = dict(json_data=json_data, meta=meta, json_path=json_path)

    def capabilities(self):
        c = super().capabilities()
        c[SERVER_CAPABILITY_EDIT_DATASET] = False
        c[SERVER_CAPABILITY_ADD_DATASET] = False
        c[SERVER_CAPABILITY_DELETE_DATASET] = False
        return c

    def __delete_entity(self, dataset_id, entity_id, kind):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        del json_data[kind][entity_id]
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def __get_entity(self, dataset_id, entity_id, kind):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        return json_data[kind][entity_id]

    def __get_entity_list(self, dataset_id, kind):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return []
        json_data = info['json_data']
        results = []
        filters = json_data[kind]
        for key in filters:
            r = filters[key]
            r['id'] = key
            results.append(r)
        return results

    def __find_dataset_id(self, entity_id, kind):
        for dataset_id in self.dataset_to_info:
            info = self.dataset_to_info.get(dataset_id)
            if info is None:
                continue
            json_data = info['json_data']
            filters = json_data[kind]
            for key in filters:
                if key == entity_id:
                    return dataset_id
        raise ValueError('{} not found'.format(entity_id))

    def __upsert_entity(self, dataset_id, entity_id, kind, entity_dict):
        if entity_id is None:
            entity_id = unique_id()
        json_data = self.dataset_to_info[dataset_id]['json_data']
        entity = json_data[kind].get(entity_id)
        if entity is None:
            entity = {}
            json_data[kind][entity_id] = entity
        entity.update(entity_dict)
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return entity_id

    def user(self, email):
        return dict()

    def datasets(self, email):
        results = []
        for key in self.dataset_to_info:
            results.append(self.dataset_to_info[key]['meta'])
        return results

    def get_dataset(self, email, dataset_id, ensure_owner=False):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:  # on the fly
            result = {'id': dataset_id, 'url': dataset_id, 'name': os.path.splitext(os.path.basename(dataset_id))[0],
                      'description': ''}
        else:
            result = info['meta']
            result['id'] = dataset_id

        return result

    def category_names(self, email, dataset_id):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return {}
        json_data = info['json_data']
        categories = json_data['categories']
        results = {}
        for category_name in categories:
            category = categories[category_name]
            cat_results = {}
            results[category_name] = cat_results
            for original_value in category:
                val = category[original_value]
                cat_results[original_value] = dict(originalValue=original_value, newValue=val.get('new'),
                                                   positiveMarkers=val.get('positiveMarkers'),
                                                   negativeMarkers=val.get('negativeMarkers'),
                                                   color=val.get('color'))
        return results

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        category_entity = json_data['categories'].get(category)
        if category_entity is None:
            category_entity = {}
            json_data['categories'][category] = category_entity
        if 'newValue' in update:
            update['new'] = update.pop('newValue')
        entity = category_entity.get(original_value)
        if entity is None:
            entity = {}
        entity.update(update)
        category_entity[original_value] = entity
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        if email is not None:
            entity['email'] = email
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def get_feature_sets(self, email, dataset_id):
        info = self.dataset_to_info.get(dataset_id)
        if info is None:
            return []
        json_data = info['json_data']
        return json_data.get('markers', [])

    def delete_feature_set(self, email, dataset_id, set_id):
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data['markers']
        for i in range(len(markers)):
            if markers[i]['id'] == set_id:
                markers.pop(i)
                break
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        if set_id is None:
            set_id = unique_id()
        else:
            self.delete_feature_set(email=email, dataset_id=dataset_id, set_id=set_id)
        json_data = self.dataset_to_info[dataset_id]['json_data']
        markers = json_data.get('markers')
        if markers is None:
            markers = []
            json_data['markers'] = markers
        markers.append(dict(id=set_id, features=features, name=name, category=category))
        write_json(json_data, self.dataset_to_info[dataset_id]['json_path'])
        return set_id

    def dataset_views(self, email, dataset_id):
        return self.__get_entity_list(dataset_id=dataset_id, kind='views')

    def delete_dataset_view(self, email, view_id):
        dataset_id = self.__find_dataset_id(view_id, 'views')
        return self.__delete_entity(dataset_id=dataset_id, entity_id=view_id, kind='views')

    def get_dataset_view(self, email, view_id):
        dataset_id = self.__find_dataset_id(view_id, 'views')
        return self.__get_entity(dataset_id=dataset_id, entity_id=view_id, kind='views')

    def upsert_dataset_view(self, email, dataset_id, view_id, name, value):
        entity = {}
        if name is not None:
            entity['name'] = name
        if value is not None:
            entity['value'] = json.dumps(value)
        if email is not None:
            entity['email'] = email
        if dataset_id is not None:
            entity['dataset_id'] = dataset_id
        return self.__upsert_entity(dataset_id=dataset_id, entity_id=view_id, kind='views', entity_dict=entity)



import concurrent.futures
import json
import os

import anndata
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import scipy.sparse
from anndata import AnnData

from cirrocumulus.abstract_dataset import AbstractDataset

max_workers = min(12, pa.cpu_count())
executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)


def read_table(path, filesystem, columns=None):
    return pq.read_table(path, filesystem=filesystem, columns=columns, use_threads=False)


def read_df(path, filesystem, columns=None):
    return read_table(path, filesystem=filesystem, columns=columns)


def read_tables(paths, filesystem, columns=None):
    futures = []
    for path in paths:
        future = executor.submit(read_table, path, filesystem=filesystem, columns=columns)
        futures.append(future)
    concurrent.futures.wait(futures)
    return futures


def read_dfs(paths, filesystem, columns=None):
    futures = []
    for path in paths:
        future = executor.submit(read_df, path, filesystem=filesystem, columns=columns)
        futures.append(future)
    concurrent.futures.wait(futures)
    return futures


def get_matrix(futures, shape=None):
    data = []
    row = []
    col = []
    is_sparse = None
    for i in range(len(futures)):
        t = futures[i].result()
        if i == 0:
            is_sparse = 'index' in t.column_names
        if is_sparse:
            row.append(t.column('index').to_numpy())
            col.append(np.repeat(i, len(t)))
        data.append(t.column('value').to_numpy())
    if len(data) == 0:
        raise ValueError('No data')
    if len(row) > 0:  # sparse
        data = np.concatenate(data)
        row = np.concatenate(row)
        col = np.concatenate(col)
        # X = scipy.sparse.coo_matrix((data, (row, col)), shape=(shape[0], len(keys))).to_csc()
        # X = scipy.sparse.csr_matrix((data, (row, col)), shape=(shape[0], len(var_keys)))
        X = scipy.sparse.csc_matrix((data, (row, col)), shape=(shape[0], len(futures)))
    else:
        X = np.array(data).T
    return X


class ParquetDataset(AbstractDataset):

    def __init__(self):
        super().__init__()

    def get_suffixes(self):
        return ['parquet', 'pq', 'cpq']

    def read_data_sparse(self, filesystem, path, keys, dataset=None, schema=None):
        shape = schema['shape']
        schema_version = schema.get('version', '1.0.0')
        X = None
        adata_modules = None
        obs = None
        var = None
        obsm = {}
        var_keys = keys.pop('X', [])
        obs_keys = keys.pop('obs', [])
        basis_keys = keys.pop('basis', [])
        module_keys = keys.pop('module', [])
        if len(var_keys) > 0:
            node_path = os.path.join(path, 'X')
            paths = [node_path + '/' + key + '.parquet' for key in var_keys]
            X = get_matrix(read_tables(paths, filesystem), shape)
            var = pd.DataFrame(index=var_keys)
        if len(module_keys) > 0:
            node_path = os.path.join(path, 'X_module')
            paths = [node_path + '/' + key + '.parquet' for key in module_keys]
            module_X = get_matrix(read_tables(paths, filesystem))
            adata_modules = anndata.AnnData(X=module_X, var=pd.DataFrame(index=module_keys))
        if len(obs_keys) > 0:
            obs = pd.DataFrame()
            node_path = os.path.join(path, 'obs')
            paths = [node_path + '/' + key + '.parquet' for key in obs_keys]
            futures = read_tables(paths, filesystem, columns=['value'])
            for i in range(len(futures)):
                df = futures[i].result().to_pandas()
                obs[obs_keys[i]] = df['value']

        if len(basis_keys) > 0:
            node_path = os.path.join(path, 'obsm')
            paths = [node_path + '/' + key + '.parquet' for key in basis_keys]
            futures = read_tables(paths, filesystem)
            for i in range(len(futures)):
                table = futures[i].result()
                vals = []
                for c in table.column_names:
                    vals.append(table.column(c))
                vals = np.array(vals).T
                obsm[basis_keys[i]] = vals
                if X is None:
                    X = scipy.sparse.coo_matrix(([], ([], [])), shape=(vals.shape[0], 0))

        adata = AnnData(X=X, obs=obs, var=var, obsm=obsm)
        if adata_modules is not None:
            adata.uns['X_module'] = adata_modules
        return adata

    def read_data_dense(self, filesystem, path, keys=None, dataset=None, schema=None):
        var_keys = keys.pop('X', [])
        obs_keys = keys.pop('obs', [])
        basis = keys.pop('basis', [])
        all_keys = obs_keys + var_keys
        X = None
        obs = None
        var = None
        obsm = {}
        for key in keys.keys():
            all_keys += keys[key]

        if len(basis) > 0:
            for key in basis:
                # 2d only
                all_keys += ['{}_{}'.format(key, 1), '{}_{}'.format(key, 2)]
        df = read_table(path, filesystem=filesystem, columns=all_keys)
        if len(var_keys) > 0:
            X = df[var_keys]
        if len(obs_keys) > 0:
            obs = df[obs_keys]
        if len(basis) > 0:
            for key in basis:
                obsm[key] = df[['{}_{}'.format(key, 1), '{}_{}'.format(key, 2)]]
        return AnnData(X=X, obs=obs, var=var, obsm=obsm)

    def read_dataset(self, filesystem, path, keys=None, dataset=None, schema=None):
        if keys is None:
            keys = {}
        # path is directory
        keys = keys.copy()
        if not path.endswith('.parquet'):
            return self.read_data_sparse(filesystem, path, keys, dataset, schema)
        return self.read_data_dense(filesystem, path, keys, dataset, schema)

    def schema(self, filesystem, path):
        if path.endswith('.json') or path.endswith('.json.gz'):
            return super().schema(filesystem, path)
        elif path.endswith('.parquet'):
            metadata = None
            with filesystem.open(path, 'rb') as s:
                parquet_file = pq.ParquetFile(s)
                schema = parquet_file.schema.to_arrow_schema()
                result = {'version': '1'}
                for metadata_key in [b'pegasus']:
                    if metadata_key in schema.metadata:
                        metadata = json.loads(schema.metadata[metadata_key])
                        break
                obs = []
                obs_cat = []
                if metadata is None:
                    # obs, the obsm, then var
                    names = parquet_file.schema.names
                    section = 'obs'
                    result['embeddings'] = []
                    result['var'] = []
                    embedding_basename_to_count = {}
                    for name in names:
                        if section == 'obs' and name.startswith('X_'):
                            section = 'obsm'
                        if section == 'obs':
                            field = schema.field(name)
                            if isinstance(field.type, pa.lib.DictionaryType):
                                obs_cat.append(name)
                            else:
                                obs.append(name)
                        if section == 'obsm' and not name.startswith('X_'):
                            section = 'var'
                        if section == 'obsm':  # X_fitsne_1', 'X_fitsne_2, etc
                            basename = name[0:name.rindex('_')]
                            count = embedding_basename_to_count.get(basename, 0)
                            embedding_basename_to_count[basename] = count + 1
                        if section == 'var':
                            result['var'].append(name)
                    for name in embedding_basename_to_count.keys():
                        result['embeddings'].append(dict(name=name, dimensions=2))
                        if embedding_basename_to_count[name] == 3:
                            result['embeddings'].append(dict(name=name, dimensions=3))
                else:
                    all_obs = metadata['obs']

                    for name in all_obs:
                        field = schema.field(name)
                        if isinstance(field.type, pa.lib.DictionaryType):
                            obs_cat.append(name)
                        else:
                            obs.append(name)
                    result['var'] = metadata['var']
                    result['embeddings'] = metadata['obsm']

                result['obs'] = obs
                result['obsCat'] = obs_cat
                result['shape'] = (parquet_file.metadata.num_rows, len(result['var']))
                return result
        else:  # directory
            return super().schema(filesystem, os.path.join(path, 'index.json.gz'))



# Abstract database class supporting datasets, categories, views, feature sets, users, jobs, and job results
# Only methods server and datasets must be supported in 'client' mode
from cirrocumulus.envir import *
from cirrocumulus.io_util import unique_id


class AbstractDB:

    def __init__(self):
        """Initializes the object

        """
        self.job_id_to_job = {}

    def capabilities(self):  # allow everything
        c = {}
        c[SERVER_CAPABILITY_RENAME_CATEGORIES] = True
        c[SERVER_CAPABILITY_JOBS] = True
        c[SERVER_CAPABILITY_SAVE_FEATURE_SETS] = True
        c[SERVER_CAPABILITY_SAVE_LINKS] = True
        c[SERVER_CAPABILITY_EDIT_DATASET] = True
        c[SERVER_CAPABILITY_ADD_DATASET] = True
        c[SERVER_CAPABILITY_DELETE_DATASET] = True
        return c

    def datasets(self, email):
        """ Gets list of available datasets

        Args:
            email: User email or None

        Returns:
            A list of dicts. Example:
            [{"id": "dataset_id",
             "name": "dataset_name"
             "title": "dataset_title",
             "owner": false,
             "url": "gs://xx/my_dataset".
             "species": "Mus Musculus"
            }]
        """
        raise NotImplementedError()

    def category_names(self, email, dataset_id):
        """Gets a dict of renamed categories for a dataset

        Args:
              email: User email or None
              dataset_id: Dataset id

        Returns:
            A dict that maps category->category_value->{}. Example:

            {"louvain":{"1":{"color":"red"}}
        """
        raise NotImplementedError()

    def upsert_category_name(self, email, dataset_id, category, original_value, update):
        """ Upserts a category name.

        Args:
             email: User email or None
             category: Category in dataset (e.g. louvain)
             dataset_id: Dataset id
             original_value: Original category value (e.g. "1")
             update: Update object optionally containing color, positiveMarkers, negativeMarkers, newValue
        """
        raise NotImplementedError()

    def user(self, email):
        """ Gets metadata about a user

        Args:
           email: User email

        Returns:
            A dict with the keys id and importer. An importer can add datasets to cirrocumulus. Example:
            {"id":"user_id",
            "importer":false
            }
        """
        raise NotImplementedError()

    # views
    def dataset_views(self, email, dataset_id):
        """ Gets list of saved dataset views (saved visualization states)

       Args:
             email: User email or None
             dataset_id: Dataset id

        Returns:
            List of dicts. Example:
            [{"id": "view id",
              "name": "view name"}]
       """
        raise NotImplementedError()

    def delete_dataset_view(self, email, dataset_id, view_id):
        """ Delete a saved view

        Args:
            email: User email or None
            dataset_id: Dataset id
            view_id: View id

        """
        raise NotImplementedError()

    def get_dataset_view(self, email, view_id):
        """ Gets detailed information for a saved dataset view

        Args:
            email: User email or None
            view_id: View id

        Returns:
            List of dicts containing id and name. Example:
            [{"id": "view id",
              "name": "view name",
              "value": "JSON encoded state"
              "notes": "view notes"
              "email": "View creator email"
        """

        raise NotImplementedError()

    def upsert_dataset_view(self, email, dataset_id, view_id, name, value):
        """ Upserts a dataset view

        Args:
              email: User email or None
              dataset_id: Dataset id
              view_id: View id or None to create new view
              name: View name
              value: JSON encoded state

         Returns:
            Upserted view id
        """
        raise NotImplementedError()

    def delete_dataset(self, email, dataset_id):
        """ Deletes a dataset

        Args:
            email: User email or None
            dataset_id: Dataset id
        """
        raise NotImplementedError()

    def upsert_dataset(self, email, dataset_id, dataset_name=None, url=None, readers=None, description=None, title=None,
                       species=None):
        """ Upserts a dataset

        Args:
              email: User email or None
              dataset_id: Dataset id
              dataset_name: Name
              url: URL
              readers: List of allowed readers
              description: Description
              title: Title
              species: Species

         Returns:
            Upserted dataset id
        """
        raise NotImplementedError()

    def get_feature_sets(self, email, dataset_id):
        """ Gets saved feature sets

        Args:
              email: User email or None
              dataset_id: Dataset id

        Returns:
            List of dicts. Example:
            [{"id": "set_id",
              "category": "set category",
              "name": "set name",
              "features: ["gene1", "gene2"]}]
        """
        raise NotImplementedError()

    def delete_feature_set(self, email, dataset_id, set_id):
        """ Deletes a saved feature set

        Args:
              email: User email or None
              dataset_id: Dataset id
              set_id: Feature set id
        """
        raise NotImplementedError()

    def upsert_feature_set(self, email, dataset_id, set_id, category, name, features):
        """ Upserts a feature set

        Args:
            email: User email or None
            dataset_id: Dataset id
            set_id: Feature set id
            category: Set category
            name: Set name
            features: List of features

        Returns:
            Upserted id
        """
        raise NotImplementedError()

    def create_job(self, email, dataset_id, job_name, job_type, params):
        """ Creates a job

        Args:
         email: User email or None
         dataset_id: Dataset id
         job_name: Job name
         job_type: Job type
         params: JSON encoded job params

        Returns:
         job id
      """
        import datetime
        job_id = unique_id()
        self.job_id_to_job[job_id] = dict(id=job_id, dataset_id=dataset_id, name=job_name, type=job_type, params=params,
                                          status=None, result=None, submitted=datetime.datetime.utcnow())
        return job_id

    def get_job(self, email, job_id, return_result):
        """ Gets a job

       Args:
          email: User email or None
          job_id: Job id
          return_result: Whether to return the job result or status only

       Returns:
          The job
      """
        job = self.job_id_to_job[job_id]
        if return_result:
            return job['result']
        return dict(id=job['id'], name=job['name'], type=job['type'], status=job['status'], submitted=job['submitted'])

    def get_jobs(self, email, dataset_id):
        """ Gets a list of all jobs for a dataset.

        Args:
            email: User email or None
            dataset_id: Dataset id

        Returns:
            List of jobs
      """
        results = []
        for job in self.job_id_to_job.values():
            results.append(dict(id=job['id'], name=job['name'], type=job['type'], status=job['status'],
                                submitted=job['submitted']))
        return results

    def delete_job(self, email, job_id):
        """ Deletes a job.

      Args:
          email: User email or None
          job_id: Job id
      """
        del self.job_id_to_job[job_id]

    def update_job(self, email, job_id, status, result):
        """ Updates job info.

          Args:
              email: User email or None
              job_id: Job id
              status: Job status
              result: Job result
          """
        job = self.job_id_to_job[job_id]
        job['status'] = status
        if result is not None:
            from cirrocumulus.util import to_json
            job['result'] = to_json(result)
